@@ -28,6 +28,9 @@
 #include <QDebug>
 #include <QFile>
 #include <QProcess>
+#include <QUuid>
+
+#include <random>
 
 #include <xcb/xcb.h>
 
@@ -36,11 +39,8 @@
 
 namespace SDDM {
     XorgDisplayServer::XorgDisplayServer(Display *parent) : DisplayServer(parent) {
-        // figure out the X11 display
-        m_display = QString(":%1").arg(displayPtr()->displayId());
-
         // get auth directory
-        QString authDir = RUNTIME_DIR;
+        QString authDir = QStringLiteral(RUNTIME_DIR);
 
         // use "." as authdir in test mode
         if (daemonApp->testing())
@@ -50,7 +50,7 @@ namespace SDDM {
         QDir().mkpath(authDir);
 
         // set auth path
-        m_authPath = QString("%1/%2").arg(authDir).arg(m_display);
+        m_authPath = QStringLiteral("%1/%2").arg(authDir).arg(QUuid::createUuid().toString());
 
         // generate cookie
         std::random_device rd;
@@ -70,10 +70,6 @@ namespace SDDM {
         stop();
     }
 
-    bool XorgDisplayServer::displayExists(int number) {
-        return QFile(QString("/tmp/.X%1-lock").arg(number)).exists();
-    }
-
     const QString &XorgDisplayServer::display() const {
         return m_display;
     }
@@ -99,7 +95,7 @@ namespace SDDM {
         file_handler.open(QIODevice::WriteOnly);
         file_handler.close();
 
-        QString cmd = QString("%1 -f %2 -q").arg(mainConfig.XDisplay.XauthPath.get()).arg(file);
+        QString cmd = QStringLiteral("%1 -f %2 -q").arg(mainConfig.X11.XauthPath.get()).arg(file);
 
         // execute xauth
         FILE *fp = popen(qPrintable(cmd), "w");
@@ -120,10 +116,6 @@ namespace SDDM {
         if (m_started)
             return false;
 
-        // generate auth file
-        addCookie(m_authPath);
-        changeOwner(m_authPath);
-
         // create process
         process = new QProcess(this);
 
@@ -135,45 +127,80 @@ namespace SDDM {
 
         if (daemonApp->testing()) {
             QStringList args;
-            args << m_display << "-ac" << "-br" << "-noreset" << "-screen" << "800x600";
-            process->start("/usr/bin/Xephyr", args);
+            args << m_display << QStringLiteral("-ac") << QStringLiteral("-br") << QStringLiteral("-noreset") << QStringLiteral("-screen") << QStringLiteral("800x600");
+            process->start(mainConfig.X11.XephyrPath.get(), args);
+
+
+            // wait for display server to start
+            if (!process->waitForStarted()) {
+                // log message
+                qCritical() << "Failed to start display server process.";
+
+                // return fail
+                return false;
+            }
+            emit started();
         } else {
             // set process environment
             QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
-            env.insert("DISPLAY", m_display);
-            env.insert("XAUTHORITY", m_authPath);
-            env.insert("XCURSOR_THEME", mainConfig.Theme.CursorTheme.get());
+            env.insert(QStringLiteral("XCURSOR_THEME"), mainConfig.Theme.CursorTheme.get());
             process->setProcessEnvironment(env);
 
-            // tell the display server to notify us when we can connect
-            SignalHandler::ignoreSigusr1();
+            //create pipe for communicating with X server
+            //0 == read from X, 1== write to from X
+            int pipeFds[2];
+            if (pipe(pipeFds) != 0) {
+                qCritical("Could not create pipe to start X server");
+            }
 
             // start display server
-            QStringList args;
-            args << m_display
-                 << "-auth" << m_authPath
-                 << "-nolisten" << "tcp"
-                 << "-background" << "none"
-                 << "-noreset"
-                 << QString("vt%1").arg(displayPtr()->terminalId());
+            QStringList args = mainConfig.X11.ServerArguments.get().split(QLatin1Char(' '), QString::SkipEmptyParts);
+            args << QStringLiteral("-auth") << m_authPath
+                 << QStringLiteral("-background") << QStringLiteral("none")
+                 << QStringLiteral("-noreset")
+                 << QStringLiteral("-displayfd") << QString::number(pipeFds[1])
+                 << QStringLiteral("vt%1").arg(displayPtr()->terminalId());
             qDebug() << "Running:"
-                     << qPrintable(mainConfig.XDisplay.ServerPath.get())
-                     << qPrintable(args.join(" "));
-            process->start(mainConfig.XDisplay.ServerPath.get(), args);
-            SignalHandler::initializeSigusr1();
-            connect(DaemonApp::instance()->signalHandler(), SIGNAL(sigusr1Received()), this, SIGNAL(started()));
-        }
+                     << qPrintable(mainConfig.X11.ServerPath.get())
+                     << qPrintable(args.join(QLatin1Char(' ')));
+            process->start(mainConfig.X11.ServerPath.get(), args);
 
-        // wait for display server to start
-        if (!process->waitForStarted()) {
-            // log message
-            qCritical() << "Failed to start display server process.";
+            // wait for display server to start
+            if (!process->waitForStarted()) {
+                // log message
+                qCritical() << "Failed to start display server process.";
+
+                // return fail
+                close(pipeFds[0]);
+                return false;
+            }
+
+            // close the other side of pipe in our process, otherwise reading
+            // from it may stuck even X server exit.
+            close(pipeFds[1]);
+
+            QFile readPipe;
+
+            if (!readPipe.open(pipeFds[0], QIODevice::ReadOnly)) {
+                qCritical("Failed to open pipe to start X Server ");
+
+                close(pipeFds[0]);
+                return false;
+            }
+            QByteArray displayNumber = readPipe.readLine();
+            displayNumber.prepend(QByteArray(":"));
+            displayNumber.remove(displayNumber.size() -1, 1); //trim trailing whitespace
+            m_display = QString::fromLocal8Bit(displayNumber);
+    
+            // close our pipe
+            close(pipeFds[0]);
 
-            // return fail
-            return false;
-        }
-        if (daemonApp->testing())
             emit started();
+        }
+
+        // generate auth file
+        addCookie(m_authPath);
+        changeOwner(m_authPath);
 
         // set flag
         m_started = true;
@@ -209,6 +236,31 @@ namespace SDDM {
         // log message
         qDebug() << "Display server stopped.";
 
+        QString displayStopCommand = mainConfig.X11.DisplayStopCommand.get();
+
+        // create display setup script process
+        QProcess *displayStopScript = new QProcess();
+
+        // set process environment
+        QProcessEnvironment env;
+        env.insert(QStringLiteral("DISPLAY"), m_display);
+        env.insert(QStringLiteral("HOME"), QStringLiteral("/"));
+        env.insert(QStringLiteral("PATH"), mainConfig.Users.DefaultPath.get());
+        env.insert(QStringLiteral("SHELL"), QStringLiteral("/bin/sh"));
+        displayStopScript->setProcessEnvironment(env);
+
+        // start display setup script
+        qDebug() << "Running display stop script " << displayStopCommand;
+        displayStopScript->start(displayStopCommand);
+
+        // wait for finished
+        if (!displayStopScript->waitForFinished(5000))
+            displayStopScript->kill();
+
+        // clean up the script process
+        displayStopScript->deleteLater();
+        displayStopScript = nullptr;
+
         // clean up
         process->deleteLater();
         process = nullptr;
@@ -221,23 +273,33 @@ namespace SDDM {
     }
 
     void XorgDisplayServer::setupDisplay() {
-        QString displayCommand = mainConfig.XDisplay.DisplayCommand.get();
+        QString displayCommand = mainConfig.X11.DisplayCommand.get();
 
         // create display setup script process
         QProcess *displayScript = new QProcess();
 
         // set process environment
         QProcessEnvironment env;
-        env.insert("DISPLAY", m_display);
-        env.insert("HOME", "/");
-        env.insert("PATH", mainConfig.Users.DefaultPath.get());
-        env.insert("XAUTHORITY", m_authPath);
-        env.insert("SHELL", "/bin/sh");
+        env.insert(QStringLiteral("DISPLAY"), m_display);
+        env.insert(QStringLiteral("HOME"), QStringLiteral("/"));
+        env.insert(QStringLiteral("PATH"), mainConfig.Users.DefaultPath.get());
+        env.insert(QStringLiteral("XAUTHORITY"), m_authPath);
+        env.insert(QStringLiteral("SHELL"), QStringLiteral("/bin/sh"));
         displayScript->setProcessEnvironment(env);
 
+        // delete displayScript on finish
+        connect(displayScript, SIGNAL(finished(int,QProcess::ExitStatus)), displayScript, SLOT(deleteLater()));
+
         // start display setup script
         qDebug() << "Running display setup script " << displayCommand;
         displayScript->start(displayCommand);
+
+        // wait for finished
+        if (!displayScript->waitForFinished(30000))
+            displayScript->kill();
+
+        // reload config if needed
+        mainConfig.load();
     }
 
     void XorgDisplayServer::changeOwner(const QString &fileName) {
